SystemOrganization addCategory: #'GrandMa-Core'!

Object subclass: #GMQueue
	instanceVariableNames: 'tasks pendingTasks mutex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GrandMa-Core'!

!GMQueue commentStamp: '<historical>' prior: 0!
a GMQueue is a FIFO of tasks.

tasks : contains the tasks that are waiting to be processed.

pendingTasks: tasks that have been consumed by a worker but do not have yet been completed.

usage:
q := GMQueue new.
q addTask: aTask.
q consumeTask. "aTask is remove from the task queue but is added to pendingTask"

q success: 'uuid of aTask'. "will notify the Queue that the job performed well."!

!GMQueue methodsFor: 'tasks' stamp: 'FrancoisStephany 4/11/2012 23:34'!
addTask: aTask
	mutex critical: [ self tasks addLast: aTask ].! !

!GMQueue methodsFor: 'tasks' stamp: 'FrancoisStephany 4/14/2012 15:06'!
consumeTask
	| task |
	^mutex critical: [
		^self tasks isEmpty
			ifTrue: [ ^nil]
			ifFalse: [
				task := self tasks removeFirst.
				task consumeDate: DateAndTime now.
				pendingTasks at: task uuid put: task]].! !

!GMQueue methodsFor: 'accessing' stamp: 'FrancoisStephany 4/12/2012 08:10'!
initialize
	tasks := LinkedList new.
	pendingTasks := Dictionary new.
	mutex := Semaphore forMutualExclusion.! !

!GMQueue methodsFor: 'accessing' stamp: 'FrancoisStephany 4/12/2012 08:10'!
pendingTasks
	^ pendingTasks! !

!GMQueue methodsFor: 'callbacks' stamp: 'FrancoisStephany 4/13/2012 15:01'!
success: aDictionary
	| uuid message task |
	uuid := aDictionary at: #uuid.
	message := aDictionary at: #message.
	task := pendingTasks at: uuid.
	task callback value: message. "relay the output to the consumer"
	self pendingTasks removeKey: uuid ifAbsent: [^self].! !

!GMQueue methodsFor: 'accessing' stamp: 'FrancoisStephany 4/11/2012 23:23'!
tasks
	^tasks! !

Object subclass: #GMServer
	instanceVariableNames: 'queue zincServer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GrandMa-Core'!

GMServer class
	instanceVariableNames: 'Current'!

!GMServer commentStamp: '<historical>' prior: 0!
a GMServer hosts a GMQueue. Adding and consuming tasks is done with HTTP.
The easiest way to get started is to use the singleton (see the testSingleton method in the tests).

Send a post request to: localhost:7777/push
with POST params:
- uuid : the UUID of the task.
- command: the command to execute.

Send a post request to: localhost:7777/pop
with POST params:
- an id string. This string is used to identify a worker image.

When popping, you will get the following JSON object:
- uuid
- command

When a consumer has executed a task, it must send a notification to the server by sending a POST request to localhost:7777/success with the following POST params:
- uuid (uuid of the task)
- result (output of the command).

use localhost:7777/failure with the following params for a failure:
- uuid (uuid of the task)
- reason (reason of the failure).!

GMServer class
	instanceVariableNames: 'Current'!

!GMServer class methodsFor: 'accessing' stamp: 'FrancoisStephany 4/13/2012 16:45'!
current
	^Current ifNil: [ Current := self startOn: self defaultPort ].! !

!GMServer class methodsFor: 'accessing' stamp: 'FrancoisStephany 4/13/2012 16:45'!
defaultPort
	^7070! !

!GMServer class methodsFor: 'accessing' stamp: 'FrancoisStephany 4/14/2012 13:48'!
reset
	Current isNil ifFalse: [ Current stop ].
	Current := nil.! !

!GMServer class methodsFor: 'instance creation' stamp: 'FrancoisStephany 4/14/2012 13:43'!
startOn: portNumber
	^self new
		zincServer: (ZnManagingMultiThreadedServer startOn: portNumber);
		yourself.! !

!GMServer methodsFor: 'delegate' stamp: 'FrancoisStephany 4/14/2012 13:51'!
buildDelegate
	^ZnValueDelegate with: [ :request | self handleRequest: request].! !

!GMServer methodsFor: 'delegate' stamp: 'FrancoisStephany 4/15/2012 15:44'!
handlePop: request
	| values uuid command task taskDictionary |
	[values := Json readFrom: request contents readStream ] 
		on: Error
		do: [ ZnResponse badRequest: request ].

	task := self queue consumeTask.
	
	"204 No Content"
	task isNil ifTrue: [
		^ZnResponse new
			statusLine: (ZnStatusLine code: 204);
			headers: ZnHeaders defaultResponseHeaders;
			yourself ].
	
	taskDictionary := Dictionary new
		at: #uuid put: task uuid;
		at: #command put: task command;
		yourself.
		
	^ZnResponse ok: (ZnEntity text: taskDictionary asJsonString).! !

!GMServer methodsFor: 'delegate' stamp: 'FrancoisStephany 4/14/2012 14:54'!
handlePush: request
	| values uuid command |
	[values := Json readFrom: request contents readStream ] 
		on: Error
		do: [ ZnResponse badRequest: request ].
		
	uuid := values at: 'uuid'.
	command := values at: 'command'.
	
	self queue addTask: (GMTask new 
		uuid: 'uuid'; 
		command: 'command';
		yourself ).
		
	^ZnResponse ok: (ZnEntity text: 'OK').! !

!GMServer methodsFor: 'delegate' stamp: 'FrancoisStephany 4/15/2012 11:54'!
handleRequest: request
	| url action |
	url := request uri.
	request method = #POST
		ifFalse: [ ^ZnResponse methodNotAllowed: request ].
	url segments size > 1
		ifTrue: [ ^ZnResponse notFound: url ].
	action := url segments first.
	action = 'push'
		ifTrue: [ ^self handlePush: request ].
	action = 'pop'
		ifTrue: [ ^self handlePop: request ].
	action = 'success'
		ifTrue: [ ^self handleSuccess: request ].
	
	^ZnResponse notFound:  url! !

!GMServer methodsFor: 'delegate' stamp: 'FrancoisStephany 4/15/2012 11:55'!
handleSuccess: request
	| values uuid resultString task |
	[values := Json readFrom: request contents readStream ] 
		on: Error
		do: [ ZnResponse badRequest: request ].
		
	uuid := values at: 'uuid'.
	resultString := values at: 'resultString'.
	
	task := self queue pendingTasks removeKey: uuid ifAbsent: [ ZnResponse badRequest: request ].
	task successCallback value: resultString.
	^ZnResponse ok: (ZnEntity text: 'OK').! !

!GMServer methodsFor: 'initialize-release' stamp: 'FrancoisStephany 4/13/2012 16:41'!
initialize
	super initialize.
	queue := GMQueue new.! !

!GMServer methodsFor: 'server' stamp: 'FrancoisStephany 4/13/2012 16:54'!
isRunning
	^zincServer isRunning.! !

!GMServer methodsFor: 'accessing' stamp: 'FrancoisStephany 4/13/2012 16:40'!
queue
	^ queue! !

!GMServer methodsFor: 'accessing' stamp: 'FrancoisStephany 4/13/2012 16:40'!
queue: anObject
	queue := anObject! !

!GMServer methodsFor: 'server' stamp: 'FrancoisStephany 4/13/2012 17:36'!
start
	self zincServer start! !

!GMServer methodsFor: 'server' stamp: 'FrancoisStephany 4/13/2012 16:47'!
stop
	self zincServer stop.! !

!GMServer methodsFor: 'accessing' stamp: 'FrancoisStephany 4/13/2012 16:40'!
zincServer
	^ zincServer! !

!GMServer methodsFor: 'accessing' stamp: 'FrancoisStephany 4/14/2012 13:47'!
zincServer: aZnServer
	zincServer isNil ifFalse: [ zincServer stop].
	zincServer := aZnServer.
	aZnServer delegate: self buildDelegate.! !

Object subclass: #GMTask
	instanceVariableNames: 'uuid command successCallback consumeDate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GrandMa-Core'!

!GMTask commentStamp: '<historical>' prior: 0!
a task has four fields:

- an uuid
- a string
- a success url to ping when the task is done.
- a failure url to ping when the task has an error.!

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/11/2012 23:59'!
command
	^ command! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/11/2012 23:59'!
command: anObject
	command := anObject! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/14/2012 15:06'!
consumeDate
	^ consumeDate! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/14/2012 15:06'!
consumeDate: aDateAndTime
	consumeDate := aDateAndTime! !

!GMTask methodsFor: 'initialize-release' stamp: 'FrancoisStephany 4/12/2012 07:49'!
initialize
	super initialize.
	uuid := UUID new printString.! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/15/2012 11:51'!
successCallback
	^ successCallback ifNil: [ successCallback := [:message | ]].! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/15/2012 11:51'!
successCallback: aBlock 
	successCallback := aBlock! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/11/2012 23:59'!
uuid
	^ uuid! !

!GMTask methodsFor: 'accessing' stamp: 'FrancoisStephany 4/14/2012 14:55'!
uuid: anUUID
	uuid := anUUID! !
